# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

# Copyright 2022 Oxide Computer Company

import copy
import json
from duro_rest.duro_client import Client, Cache

class FlattenedBOM(list):
  def __init__(self, data: list):
    list.__init__(self, data)
    self.__data = data

  def __eq__(self, other):
    if isinstance(other, FlattenedBOM):
        return self.__data == other.inner()

    return False

  def inner(self) -> list:
    return self.__data

class IndentedBOM(list):
  def __init__(self, data: list):
    list.__init__(self, data)
    self.__data = data

  def __eq__(self, other):
    if isinstance(other, IndentedBOM):
        return self.__data == other.inner()

    return False

  def inner(self) -> list:
    return self.__data

  def flattened(self) -> FlattenedBOM:
    source = copy.deepcopy(self.__data)

    # Create a list to hold the unique components in the BOM
    unique_components = []

    # Keys that are allowed to differ between components. We will either reconcile these values (as
    # we do in the case of quantity and refDes) or we may drop / ignore the value (as we do in the
    # case of itemNumber)
    allowed_conflicts = [
      'quantity',                 # Sum
      'refDes'                    # Concat
    ]

    for component in source:
      # For a flattened BOM we are only interested in the multiplied quantities
      component['quantity'] = component['multiplied_quantity']

      # We need to drop a few fields that we are not interested in, or can not merge
      component.pop('isAddedAfterPullRequest', None)
      component.pop('itemNumber', None)
      component.pop('multiplied_quantity', None)
      component.pop('parents', None)
      component.pop('parent_cpns', None)
      component.pop('level', None)

      # For each component, check the id to see if we already have it in the list of unique
      # components
      existing = next((c for c in unique_components if component['_id'] == c['_id']), None)

      # If we found an existing component, sum together the quantities of the current component and
      # the existing component
      if existing:

        # Check that the two components differ at most in terms of quantity
        for k, v in component.items():
          assert k in allowed_conflicts or existing[k] == v, "Values for key `" + k + "` do not match. Unable to merge components. Existing (" + str(existing[k]) + ") != Current (" + str(v) + ")"

        existing['quantity'] = existing['quantity'] + component['quantity']

        if 'refDes' in component and component['refDes'] != '':
          existing['refDes'] = existing['refDes'] + ',' + component['refDes']
      else:
        # Otherwise this is a component we have not seen yet. Just add it to the list
        unique_components.append(component)

    return FlattenedBOM(unique_components)

class NestedBOM(dict):
  def __init__(self, data: dict):
    # Assign a default quantity of 1 to the top line item if it has not yet been assigned a quantity
    if not 'quantity' in data:
      data['quantity'] = 1

    dict.__init__(self, data)
    self.__data = data

  def __eq__(self, other):
    if isinstance(other, NestedBOM):
        return self.__data == other.inner()

    return False

  def inner(self) -> dict:
    return self.__data

  def indented(self) -> IndentedBOM:
    source = copy.deepcopy(self.__data)

    # We need to first take the nested BOM and transform it into a flat list of components. The
    # quantity for a given component is multiplied by each of its parents quantity. This list will
    # contain duplicate components which will need to be deduplicated when transforming to a
    # flattened BOM
    components = component_list(source)

    return IndentedBOM(components)

def component_list(component) -> list:
  # When we start processing a component, set its multiplied quantity to its non-multiplied
  # quantity. As we traverse back up the tree this field will be updated as needed
  component['multiplied_quantity'] = component['quantity']

  # Start an empty lit of parent component ids
  component['parents'] = []
  component['parent_cpns'] = []

  # Component depth starts at 0 and will be incremented as we walk back up the tree
  component['level'] = 0

  # The component passed in represents a portion of the nested BOM. Here we are going to flatten all
  # of this component's children (and nested children) by calling component_list on each child
  # component. Since component_list returns a list of components, the map call here will end up
  # with a list of lists. Therefore we call flatten to turn that result from a list of lists into
  # a list of components
  flattened_children = flatten(map(component_list, component['children']))

  for child in flattened_children:
    # For each child component we multiply its quantity by the quantity of the current component
    # we are looking at. We store the value to make future work easier
    child['multiplied_quantity'] = component['multiplied_quantity'] * child['multiplied_quantity']

    # We also want to track the parents list to identify the path back up the tree from a leaf
    child['parents'].append(component['_id'])
    child['parent_cpns'].append(component['cpn'])

    # Increment the level of each child component by one
    child['level'] = child['level'] + 1

  # Since we are flattening the BOM we no longer want the components to have references to their
  # children
  component.pop('children', None)

  # Finally return a flattened list that contains the current component and all of its children
  return [component] + flattened_children

class BOMClient(Client):
  def __init__(self, api_key, url = "https://public-api.duro.app", cache = Cache(0)):
    super().__init__(api_key, url, cache)

  def product_bom(self, id) -> NestedBOM:
    return self.__nested_bom(self.product(id))

  def component_bom(self, id) -> NestedBOM:
    return self.__nested_bom(self.component(id))

  def __nested_bom(self, parent) -> NestedBOM:
    components = self.components()

    # We we request a product or component from Duro the 'children' property only contains a single
    # level of depth. What we need is the full nested BOM. We call expand_child_components on each
    # child to build out that full tree.
    parent['children'] = list(map(lambda child: expand_child_components(child, components), parent['children']))

    return NestedBOM(parent)

def expand_child_components(parent, components):
  # The components list contains most of the needed data, but we need some of the properties that
  # are defined by the parent component (specifically the quantity). Therefore we combine the
  # properties of the parent we are currently expanding with the expanded data available from
  # fetching all of the components
  matching_components = list(component for component in components if component['_id'] == parent['component'])

  if len(matching_components) == 0:
    raise LookupError('Failed to find component in cache ' + parent['component'])
  elif len(matching_components) > 1:
    raise LookupError('Found multiple components in cache ' + parent['component'])

  expanded_parent = {**parent, **matching_components[0]}

  # Now that we have expanded the parent component, we need to expand each of the child components
  expanded_parent['children'] = list(map(lambda child: expand_child_components(child, components), expanded_parent['children']))

  return expanded_parent

def flatten(l):
  return [item for sublist in l for item in sublist]


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['duro_rest']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2,<3']

setup_kwargs = {
    'name': 'duro-rest',
    'version': '0.0.7',
    'description': 'An API client for the Duro REST API',
    'long_description': '# duro_rest\n\nAn API client for the [Duro REST API](https://public-api.duro.app/v1/docs/).\n\n## Installation\n\nInstalling from [PyPi]()\n\n```pip install duro_rest```\n\nInstalling from [conda-forge]()\n\nTBD\n\n## Examples\n\nThe Duro client is a very minimal API client for reading data from the [Duro v1 REST API](https://public-api.duro.app/v1/docs/). It does not define classes for each object type and instead returns plain dictionaries\nand lists.\n\n#### Base Client\n\n```python\nfrom duro_rest import Client\n\n# Create a new client, providing your API key. By default clients will make calls to the public REST\n# API, but the endpoint base can be overridden if needed\nclient = Client("your-api-key")\n\n# Fetch a component by its Duro assigned id (as opposed to the CPN)\ncomponent = client.component("component-id")\n\n# Fetch a list of all components in Duro\ncomponents = client.components()\n\n# Fetch a list of all components in Duro filtered by status\ncomponents = client.components(status = "OBSOLETE")\n```\n\n#### BOM Client\n\nThe Duro API does not support pulling BOMs directly. The BOM Client is here to make that experience\na little easier.\n\n```python\nfrom duro_rest import BOMClient\n\n# Create a new bom client for fetching full BOMs (either nested or flattened). It accepts the same\n# arguments as the base client\nclient = Client("your-api-key")\n\n# Get a nested BOM starting from a product\nnested_product_bom = client.product_bom("product-id")\n\n# Get a nested BOM starting from a component\nnested_component_bom = client.component("component-id")\n\n# Convert the nested BOM into flattened BOM\nindented_product_bom = nested_product_bom.idented()\n\n# Flattening will collapse rows for the same components together\nflattened_product_bom = indented_product_bom.flatten()\n```',
    'author': 'Augustus Mayo',
    'author_email': 'augustus@oxidecomputer.com',
    'maintainer': 'Augustus Mayo',
    'maintainer_email': 'augustus@oxidecomputer.com',
    'url': 'https://pypi.org/project/duro-rest/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

# Copyright 2022 Oxide Computer Company

from datetime import datetime, timedelta
import json
from json import JSONDecodeError
import math
import requests

from duro_rest.error import AuthenticationError, BadRequestError, ForbiddenError, NotFoundError

class Cache():
  def __init__(self, cache_duration):
    self.__delta = timedelta(seconds = cache_duration)
    self.__data = {}

  def get(self, key):
    if key in self.__data:
      (value, expiration) = self.__data[key]

      if datetime.now().timestamp() < expiration:
        return value

    return None

  def insert(self, key, value):
    if self.__delta.seconds > 0:
      self.__data[key] = (value, (datetime.now() + self.__delta).timestamp())

class FileCache(Cache):
  def __init__(self, cache_duration, path):
    super().__init__(cache_duration)

    try:
      self.__cache_file = open(path, "r+")
      stored = json.load(self.__cache_file)

      if stored != None:
        self._Cache__data = stored
    except TypeError:
      # Ignore
      pass
    except JSONDecodeError:
      # Ignore
      pass
    except FileNotFoundError:
      self.__cache_file = open(path, "w")
      pass
    finally:
      # Make sure we always rewind to the start of the file
      self.__cache_file.seek(0, 0)

  def save(self):
    self.__cache_file.truncate()
    self.__cache_file.write(json.dumps(self._Cache__data))
    self.__cache_file.seek(0, 0)

class Client:
  def __init__(self, api_key, url = "https://public-api.duro.app", cache = Cache(0)):
    self.__base_url = url + "/v1/"
    self.__api_key = api_key
    self.__cache = cache

  def categories(
    self,
    name = None,
    type_ = None,
    prefix = None
  ):
    return self.__get_paginated(
      "categories",
      "categories",
      {
        "name": name,
        "type": type_,
        "prefix": prefix
      }
    )

  def change_order(self, id):
    return self.__get("changeorders/" + id)

  def change_orders(
    self,
    con = None,
    name = None,
    status = None,
    resolution = None
  ):
    return self.__get_paginated(
      "changeorders",
      "changeOrders",
      {
        "con": con,
        "name": name,
        "status": status,
        "resolution": resolution
      }
    )

  def component(self, id):

    # Try to serve from the cache
    if entries := self.__cache.get("components"):
      for entry in entries:
        if entry["_id"] == id:
          return entry

    return self.__get("components/" + id)

  def components(
    self,
    mass = None,
    modified = None,
    cpn = None,
    name = None,
    eid = None,
    revision = None,
    category = None,
    cat_group = None,
    status = None,
    document_name = None,
    document_mime_type = None,
    document_revision = None,
    document_status = None,
    document_type = None,
    procurement = None
  ):
    return self.__get_paginated(
      "components",
      "components",
      {
        "mass": mass,
        "modified": modified,
        "cpn": cpn,
        "name": name,
        "eid": eid,
        "revision": revision,
        "category": category,
        "catGroup": cat_group,
        "status": status,
        "documentName": document_name,
        "documentMimeType": document_mime_type,
        "documentRevision": document_revision,
        "documentStatus": document_status,
        "documentType": document_type,
        "procurement": procurement
      }
    )

  def component_revision(self, id):
    return self.__get("component/revision/" + id)

  def component_revisions(
    self,
    mass = None,
    modified = None,
    cpn = None,
    name = None,
    eid = None,
    revision = None,
    category = None,
    cat_group = None,
    status = None,
    document_name = None,
    document_mime_type = None,
    document_revision = None,
    document_status = None,
    document_type = None,
    procurement = None
  ):
    return self.__get_paginated(
      "component/revision",
      "componentRevisions",
      {
        "mass": mass,
        "modified": modified,
        "cpn": cpn,
        "name": name,
        "eid": eid,
        "revision": revision,
        "category": category,
        "catGroup": cat_group,
        "status": status,
        "documentName": document_name,
        "documentMimeType": document_mime_type,
        "documentRevision": document_revision,
        "documentStatus": document_status,
        "documentType": document_type,
        "procurement": procurement
      }
    )

  def document(self, id):
    return self.__get("documents/" + id)

  def product(self, id):
    return self.__get("products/" + id)

  def products(
    self,
    mass = None,
    modified = None,
    cpn = None,
    name = None,
    eid = None,
    revision = None,
    status = None,
    document_name = None,
    document_mime_type = None,
    document_revision = None,
    document_status = None,
    document_type = None,
    procurement = None
  ):
    return self.__get_paginated(
      "products",
      "products",
      {
        "mass": mass,
        "modified": modified,
        "cpn": cpn,
        "name": name,
        "eid": eid,
        "revision": revision,
        "status": status,
        "documentName": document_name,
        "documentMimeType": document_mime_type,
        "documentRevision": document_revision,
        "documentStatus": document_status,
        "documentType": document_type,
        "procurement": procurement
      }
    )

  def product_revision(self, id):
    return self.__get("product/revision/" + id)

  def product_revisions(
    self,
    mass = None,
    modified = None,
    cpn = None,
    name = None,
    eid = None,
    revision = None,
    status = None,
    document_name = None,
    document_mime_type = None,
    document_revision = None,
    document_status = None,
    document_type = None,
    procurement = None
  ):
    return self.__get_paginated(
      "product/revision",
      "productRevisions",
      {
        "mass": mass,
        "modified": modified,
        "cpn": cpn,
        "name": name,
        "eid": eid,
        "revision": revision,
        "status": status,
        "documentName": document_name,
        "documentMimeType": document_mime_type,
        "documentRevision": document_revision,
        "documentStatus": document_status,
        "documentType": document_type,
        "procurement": procurement
      }
    )

  def user(self, id):
    return self.__get("users/" + id)

  def users(
    self,
    email = None,
    role = None
  ):
    return self.__get_paginated(
      "users",
      "users",
      {
        "email": email,
        "role": role
      }
    )

  def __get_paginated(self, path, resultKey, params):
    resources = []

    page = self.__get(path, params = {"perPage": 100, **params})

    total_results = page["resultCount"]
    pages = math.ceil(total_results / 100)

    resources = resources + page[resultKey]

    if pages > 1:
      for page_index in range(2, pages + 1):
        page = self.__get(path, params = {**params, "perPage": 100, "page": page_index})
        resources = resources + page[resultKey]

    # For each resource we found populate its individual record in the cache
    for resource in resources:
      self.__cache.insert(path + "/" + resource["_id"] + "::{}", resource)

    return resources

  def __get(self, path, headers = {}, params = {}, data = None):
    cache_key = path + "::" + json.dumps(params)

    if (value := self.__cache.get(cache_key)) is not None:
      return value

    headers["x-api-key"] = self.__api_key

    resp = requests.get(self.__base_url + path, headers = headers, params = params, data = data)

    if resp.status_code == 400:
      raise BadRequestError(resp.json()["message"], resp.json()["errors"])
    if resp.status_code == 401:
      raise AuthenticationError(resp.json()["message"])
    if resp.status_code == 403:
      raise ForbiddenError(resp.json()["message"])
    if resp.status_code == 404:
      raise NotFoundError(resp.json()["message"])
    else:
      data = resp.json()
      self.__cache.insert(cache_key, data)
      return data